import autofit as af
import autolens as al

"""
In this pipeline, we fit the a strong lens using an _EllipticalIsothermal_ _MassProfile_ and a source composed of 4 
parametric _EllipticalSersic_'s.

The pipeline is four phases:

Phase 1:

    Fit the _EllipticalIsothermal_ mass model and the first _EllipticalSersic_ light profile of the source.

    Lens Light: None
    Lens Mass: None
    Source Light: None
    Prior Passing: None
    Notes: None

Phase 2:

    Add the second _EllipticalSersic_ to the source model.

    Lens Light: None
    Lens Mass: EllipticalIsothermal
    Source Light: _EllipticalSersic_ + EllipticalSersic
    Prior Passing: Lens Mass (model -> phase 1), Source Light (model -> phase 1).
    Notes: Uses the previous mass model and source model to initialize the non-linear search.

Phase 3:

    Add the third _EllipticalSersic_ to the source model.

    Lens Light: None
    Lens Mass: EllipticalIsothermal
    Source Light: _EllipticalSersic_ + _EllipticalSersic_ + EllipticalSersic
    Prior Passing: Lens Mass (model -> phase 2), Source Light (model -> phase 2).
    Notes: Uses the previous mass model and source model to initialize the non-linear search.

Phase 4:

    Add the fourth _EllipticalSersic_ to the source model.

    Lens Light: None
    Lens Mass: EllipticalIsothermal
    Source Light: _EllipticalSersic_ + _EllipticalSersic_ + _EllipticalSersic_ + EllipticalSersic
    Prior Passing: Lens Mass (model -> phase 3), Source Light (model -> phase 3).
    Notes: Uses the previous mass model and source model to initialize the non-linear search.

"""


def make_pipeline(setup, settings):

    """SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS"""

    pipeline_name = "pipeline__complex_source"

    setup.folders.append(pipeline_name)
    setup.folders.append(setup.tag)

    """
    Phase 1: Initialize the lens's mass model using a simple source.
    
    This won't fit the complicated structure of the source, but it'll give us a reasonable estimate of the
    einstein radius and the other lens-mass parameters.
    """

    phase1 = al.PhaseImaging(
        phase_name="phase_1__mass_sie__source_x1_sersic",
        folders=setup.folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, mass=al.mp.EllipticalIsothermal),
            source=al.GalaxyModel(redshift=1.0, sersic_0=al.lp.EllipticalSersic),
        ),
        settings=settings,
        search=af.DynestyStatic(n_live_points=40, evidence_tolerance=5.0),
    )

    """
    Phase 1: Add a second source component, using the previous model as the initialization on the lens / source
             parameters. We'll vary the parameters of the lens mass model and first source galaxy component during the 
             fit.
    """

    phase2 = al.PhaseImaging(
        phase_name="phase_2__mass_sie__source_sersic_x2",
        folders=setup.folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase1.result.model.galaxies.lens.mass
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                sersic_0=phase1.result.model.galaxies.source.sersic_0,
                sersic_1=al.lp.EllipticalSersic,
            ),
        ),
        settings=settings,
        search=af.DynestyStatic(n_live_points=40, evidence_tolerance=5.0),
    )

    """Phase 3: Same again, but with 3 source galaxy components."""

    phase3 = al.PhaseImaging(
        phase_name="phase_3__mass_sie__source_sersic_x3",
        folders=setup.folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase2.result.model.galaxies.lens.mass
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                sersic_0=phase2.result.model.galaxies.source.sersic_0,
                sersic_1=phase2.result.model.galaxies.source.sersic_1,
                sersic_2=al.lp.EllipticalSersic,
            ),
        ),
        settings=settings,
        search=af.DynestyStatic(n_live_points=50, evidence_tolerance=5.0),
    )

    """Phase 4: And one more for luck!"""

    phase4 = al.PhaseImaging(
        phase_name="phase_4__mass_sie__source_sersic_x4",
        folders=setup.folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=phase3.result.model.galaxies.lens.mass
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                sersic_0=phase3.result.model.galaxies.source.sersic_0,
                sersic_1=phase3.result.model.galaxies.source.sersic_1,
                sersic_2=phase3.result.model.galaxies.source.sersic_2,
                sersic_3=al.lp.EllipticalSersic,
            ),
        ),
        settings=settings,
        search=af.DynestyStatic(n_live_points=50, evidence_tolerance=0.3),
    )

    return al.PipelineDataset(pipeline_name, phase1, phase2, phase3, phase4)



import autofit as af
import autolens as al


class TestSLaMPipelineSource:
    def test__shear(self):

        setup_mass = al.SetupMassTotal(no_shear=False)
        pipeline_source = al.SLaMPipelineSourceParametric(setup_mass=setup_mass)

        assert pipeline_source.shear is al.mp.ExternalShear

        setup_mass = al.SetupMassTotal(no_shear=True)
        pipeline_source = al.SLaMPipelineSourceParametric(setup_mass=setup_mass)

        assert pipeline_source.shear == None


class TestSLaMPipelineMass:
    def test__fix_lens_light_tag(self):

        pipeline_mass = al.SLaMPipelineMass(fix_lens_light=False)
        assert pipeline_mass.fix_lens_light_tag == ""
        pipeline_mass = al.SLaMPipelineMass(fix_lens_light=True)
        assert pipeline_mass.fix_lens_light_tag == "__fix_lens_light"

    def test__shear_from_previous_pipeline(self):

        setup_mass = al.SetupMassTotal(no_shear=True)
        pipeline_mass = al.SLaMPipelineMass(setup_mass=setup_mass)

        assert pipeline_mass.shear_from_previous_pipeline == None

        setup_mass = al.SetupMassTotal(no_shear=False)
        pipeline_mass = al.SLaMPipelineMass(setup_mass=setup_mass)

        assert isinstance(
            pipeline_mass.shear_from_previous_pipeline, af.AbstractPromise
        )



from autoconf import conf
from autoarray.inversion import pixelizations as pix, inversions as inv
from autogalaxy.dataset import imaging, interferometer
from autogalaxy.pipeline.phase import settings
from autolens.lens.settings import SettingsLens


class SettingsPhaseImaging(settings.SettingsPhaseImaging):
    def __init__(
        self,
        settings_masked_imaging=imaging.SettingsMaskedImaging(),
        settings_pixelization=pix.SettingsPixelization(use_border=True),
        settings_inversion=inv.SettingsInversion(),
        settings_lens=SettingsLens(),
        log_likelihood_cap=None,
    ):

        super().__init__(
            settings_masked_imaging=settings_masked_imaging,
            settings_pixelization=settings_pixelization,
            settings_inversion=settings_inversion,
            log_likelihood_cap=log_likelihood_cap,
        )

        self.settings_lens = settings_lens

    @property
    def phase_tag_no_inversion(self):

        return (
            f"{conf.instance.settings_tag.get('phase', 'phase')}__"
            f"{self.settings_masked_imaging.tag_no_inversion}__"
            f"{self.settings_lens.tag}"
            f"{self.log_likelihood_cap_tag}"
        )

    @property
    def phase_tag_with_inversion(self):
        return (
            f"{conf.instance.settings_tag.get('phase', 'phase')}__"
            f"{self.settings_masked_imaging.tag_with_inversion}__"
            f"{self.settings_lens.tag}__"
            f"{self.settings_pixelization.tag}__"
            f"{self.settings_inversion.tag}"
            f"{self.log_likelihood_cap_tag}"
        )


class SettingsPhaseInterferometer(settings.SettingsPhaseInterferometer):
    def __init__(
        self,
        masked_interferometer=interferometer.SettingsMaskedInterferometer(),
        settings_pixelization=pix.SettingsPixelization(use_border=True),
        settings_inversion=inv.SettingsInversion(),
        settings_lens=SettingsLens(),
        log_likelihood_cap=None,
    ):

        super().__init__(
            masked_interferometer=masked_interferometer,
            settings_pixelization=settings_pixelization,
            settings_inversion=settings_inversion,
            log_likelihood_cap=log_likelihood_cap,
        )

        self.settings_lens = settings_lens

    @property
    def phase_tag_no_inversion(self):
        return (
            f"{conf.instance.settings_tag.get('phase', 'phase')}__"
            f"{self.settings_masked_interferometer.tag_no_inversion}__"
            f"{self.settings_lens.tag}"
            f"{self.log_likelihood_cap_tag}"
        )

    @property
    def phase_tag_with_inversion(self):

        return (
            f"{conf.instance.settings_tag.get('phase', 'phase')}__"
            f"{self.settings_masked_interferometer.tag_with_inversion}__"
            f"{self.settings_lens.tag}__"
            f"{self.settings_pixelization.tag}__"
            f"{self.settings_inversion.tag}"
            f"{self.log_likelihood_cap_tag}"
        )


class SettingsPhasePositions:
    def __init__(self, lens):

        self.lens = lens



import autolens as al


class TestSetupMass:
    def test__no_shear_tag(self):
        setup = al.SetupMassLightDark(no_shear=False)
        assert setup.no_shear_tag == "__with_shear"

        setup = al.SetupMassLightDark(no_shear=True)
        assert setup.no_shear_tag == "__no_shear"


class TestSetupSubhalo:
    def test__subhalo_centre_tag(self):

        setup = al.SetupSubhalo(subhalo_instance=None)
        assert setup.subhalo_centre_tag == ""
        setup = al.SetupSubhalo(subhalo_instance=al.mp.SphericalNFW(centre=(2.0, 2.0)))
        assert setup.subhalo_centre_tag == "__sub_centre_(2.00,2.00)"
        setup = al.SetupSubhalo(subhalo_instance=al.mp.SphericalNFW(centre=(3.0, 4.0)))
        assert setup.subhalo_centre_tag == "__sub_centre_(3.00,4.00)"
        setup = al.SetupSubhalo(
            subhalo_instance=al.mp.SphericalNFW(centre=(3.027, 4.033))
        )
        assert setup.subhalo_centre_tag == "__sub_centre_(3.03,4.03)"

    def test__subhalo_mass_at_200_tag(self):

        setup = al.SetupSubhalo(subhalo_instance=None)
        assert setup.subhalo_mass_at_200_tag == ""
        setup = al.SetupSubhalo(
            subhalo_instance=al.mp.SphericalNFWMCRLudlow(mass_at_200=1e8)
        )
        assert setup.subhalo_mass_at_200_tag == "__sub_mass_1.0e+08"
        setup = al.SetupSubhalo(
            subhalo_instance=al.mp.SphericalNFWMCRLudlow(mass_at_200=1e9)
        )
        assert setup.subhalo_mass_at_200_tag == "__sub_mass_1.0e+09"
        setup = al.SetupSubhalo(
            subhalo_instance=al.mp.SphericalNFWMCRLudlow(mass_at_200=1e10)
        )
        assert setup.subhalo_mass_at_200_tag == "__sub_mass_1.0e+10"


class TestSetupPipeline:
    def test__tag(self):

        hyper = al.SetupHyper(
            hyper_galaxies=True, hyper_background_noise=True, hyper_image_sky=True
        )

        setup_mass = al.SetupMassLightDark(align_bulge_dark_centre=True)

        setup = al.SetupPipeline(setup_hyper=hyper, setup_mass=setup_mass)

        assert (
            setup.tag == "setup__"
            "hyper[galaxies_bg_sky_bg_noise]__"
            "mass[light_dark__with_shear__mlr_free__align_bulge_dark_centre]"
        )

        setup_source = al.SetupSourceInversion(
            pixelization=al.pix.Rectangular, regularization=al.reg.Constant
        )

        setup_light = al.SetupLightBulgeDisk(light_centre=(1.0, 2.0))

        setup_mass = al.SetupMassLightDark(mass_centre=(3.0, 4.0), no_shear=True)

        setup = al.SetupPipeline(
            setup_source=setup_source, setup_light=setup_light, setup_mass=setup_mass
        )

        assert (
            setup.tag == "setup__"
            "light[bulge_disk__light_centre_(1.00,2.00)]__"
            "mass[light_dark__mass_centre_(3.00,4.00)__no_shear__mlr_free]__"
            "source[pix_rect__reg_const]"
        )

        setup_mass = al.SetupMassLightDark(align_light_dark_centre=True)

        setup = al.SetupPipeline(setup_mass=setup_mass)

        assert (
            setup.tag == "setup__"
            "mass[light_dark__with_shear__mlr_free__align_light_dark_centre]"
        )

        smbh = al.SetupSMBH(include_smbh=True, smbh_centre_fixed=True)

        subhalo = al.SetupSubhalo(
            subhalo_instance=al.mp.SphericalNFWMCRLudlow(
                centre=(1.0, 2.0), mass_at_200=1e8
            )
        )

        setup = al.SetupPipeline(setup_smbh=smbh, subhalo=subhalo)

        assert (
            setup.tag == "setup__"
            "smbh[centre_fixed]__"
            "subhalo[nfw__sub_centre_(1.00,2.00)__sub_mass_1.0e+08]"
        )



import autolens as al
import numpy as np
import pytest
from autoarray.inversion import inversions
from test_autogalaxy.mock import MockLightProfile


class TestFitImaging:
    class TestFitProperties:
        def test__total_inversions(self, masked_imaging_7x7):

            g0 = al.Galaxy(redshift=0.5)

            g1 = al.Galaxy(redshift=1.0)

            g2 = al.Galaxy(redshift=2.0)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            assert fit.total_inversions == 0

            g2 = al.Galaxy(
                redshift=2.0,
                pixelization=al.pix.Rectangular(),
                regularization=al.reg.Constant(),
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            assert fit.total_inversions == 1

            g0 = al.Galaxy(
                redshift=0.5,
                pixelization=al.pix.Rectangular(),
                regularization=al.reg.Constant(),
            )

            g1 = al.Galaxy(
                redshift=1.0,
                pixelization=al.pix.Rectangular(),
                regularization=al.reg.Constant(),
            )

            g2 = al.Galaxy(
                redshift=2.0,
                pixelization=al.pix.Rectangular(),
                regularization=al.reg.Constant(),
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            assert fit.total_inversions == 3

    class TestLikelihood:
        def test__1x2_image__no_psf_blurring__tracing_fits_data_with_chi_sq_5(self):
            # The image plane image generated by the galaxy is [1.0, 1.0]

            # Thus the chi squared is 4.0**2.0 + 3.0**2.0 = 25.0

            psf = al.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
            )

            imaging = al.Imaging(
                image=5.0 * al.Array.ones(shape_2d=(3, 4)),
                psf=psf,
                noise_map=al.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[6] = 4.0

            mask = al.Mask.manual(
                mask=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
            )

            masked_imaging_7x7 = al.MaskedImaging(
                imaging=imaging,
                mask=mask,
                settings=al.SettingsMaskedImaging(sub_size=1),
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = al.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = al.Tracer.from_galaxies(galaxies=[g0])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 9.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test__1x2_image__include_psf_blurring__tracing_fits_data_with_chi_sq_4(
            self
        ):
            # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

            # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

            psf = al.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 3.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
                renormalize=False,
            )

            imaging = al.Imaging(
                image=5.0 * al.Array.ones(shape_2d=(3, 4)),
                psf=psf,
                noise_map=al.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[6] = 4.0

            mask = al.Mask.manual(
                mask=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
            )

            masked_imaging_7x7 = al.MaskedImaging(
                imaging=imaging,
                mask=mask,
                settings=al.SettingsMaskedImaging(sub_size=1, renormalize_psf=False),
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = al.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = al.Tracer.from_galaxies(galaxies=[g0])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 16.0
            assert fit.reduced_chi_squared == 16.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                16.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test_hyper_galaxy_changes_noise_above_from_1_to_2__reflected_in_likelihood(
            self
        ):
            # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

            # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

            # The hyper_galaxies galaxy increases the noise in both pixels by 1.0, to 2.0.

            # This reduces the chi squared to 2.0 instead of 4.0

            psf = al.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 3.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
            )

            imaging = al.Imaging(
                image=5.0 * al.Array.ones(shape_2d=(3, 4)),
                psf=psf,
                noise_map=al.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[6] = 4.0

            mask = al.Mask.manual(
                mask=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
            )

            masked_imaging_7x7 = al.MaskedImaging(
                imaging=imaging,
                mask=mask,
                settings=al.SettingsMaskedImaging(sub_size=1, renormalize_psf=False),
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=MockLightProfile(value=1.0, size=2),
                hyper_galaxy=al.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image=al.Array.ones(shape_2d=(1, 2)),
                hyper_galaxy_image=al.Array.ones(shape_2d=(1, 2)),
                hyper_minimum_value=0.0,
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 2.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 4.0
            assert fit.reduced_chi_squared == 4.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                4.0 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0)
            )

        def test__hyper_image_changes_background_sky__reflected_in_likelihood(self):

            psf = al.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
            )

            imaging = al.Imaging(
                image=al.Array.full(fill_value=4.0, shape_2d=(3, 4)),
                psf=psf,
                noise_map=al.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[5] = 5.0

            mask = al.Mask.manual(
                mask=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
            )

            masked_imaging_7x7 = al.MaskedImaging(
                imaging=imaging,
                mask=mask,
                settings=al.SettingsMaskedImaging(sub_size=1),
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = al.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = al.Tracer.from_galaxies(galaxies=[g0])

            hyper_image_sky = al.hyper_data.HyperImageSky(sky_scale=1.0)

            fit = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
            )

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 6.0, 5.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 25.0, 16.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 41.0
            assert fit.reduced_chi_squared == 41.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                41.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test__hyper_background_changes_background_noise_map__reflected_in_likelihood(
            self
        ):

            psf = al.Kernel.manual_2d(
                array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
                pixel_scales=1.0,
            )

            imaging = al.Imaging(
                image=5.0 * al.Array.ones(shape_2d=(3, 4)),
                psf=psf,
                noise_map=al.Array.ones(shape_2d=(3, 4)),
            )
            imaging.image[6] = 4.0

            mask = al.Mask.manual(
                mask=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
            )

            masked_imaging_7x7 = al.MaskedImaging(
                imaging=imaging,
                mask=mask,
                settings=al.SettingsMaskedImaging(sub_size=1),
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = al.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = al.Tracer.from_galaxies(galaxies=[g0])

            hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            fit = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                hyper_background_noise=hyper_background_noise,
            )

            assert (
                fit.mask
                == np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                )
            ).all()

            assert (
                fit.image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.noise_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 2.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.model_image.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.normalized_residual_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 1.5, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert (
                fit.chi_squared_map.in_2d
                == np.array(
                    [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 2.25, 0.0], [0.0, 0.0, 0.0, 0.0]]
                )
            ).all()

            assert fit.chi_squared == 6.25
            assert fit.reduced_chi_squared == 6.25 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                6.25 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0)
            )

    class TestCompareToManualProfilesOnly:
        def test___all_lens_fit_quantities__no_hyper_methods(self, masked_imaging_7x7):

            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                mass_profile=al.mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = al.Galaxy(
                redshift=1.0, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            assert masked_imaging_7x7.noise_map.in_2d == pytest.approx(
                fit.noise_map.in_2d
            )

            model_image = tracer.blurred_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert model_image.in_2d == pytest.approx(fit.model_image.in_2d)

            residual_map = al.util.fit.residual_map_from(
                data=masked_imaging_7x7.image, model_data=model_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(fit.chi_squared_map.in_2d)

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=masked_imaging_7x7.noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)
            assert log_likelihood == fit.figure_of_merit

        def test___lens_fit_galaxy_model_image_dict__corresponds_to_blurred_galaxy_images(
            self, masked_imaging_7x7
        ):
            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                mass_profile=al.mp.SphericalIsothermal(einstein_radius=1.0),
            )
            g1 = al.Galaxy(
                redshift=1.0, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )
            g2 = al.Galaxy(redshift=1.0)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            traced_grids_of_planes = tracer.traced_grids_of_planes_from_grid(
                grid=masked_imaging_7x7.grid
            )
            traced_blurring_grids_of_planes = tracer.traced_grids_of_planes_from_grid(
                grid=masked_imaging_7x7.blurring_grid
            )

            g0_image = g0.image_from_grid(grid=traced_grids_of_planes[0])
            g0_blurring_image = g0.image_from_grid(
                grid=traced_blurring_grids_of_planes[0]
            )

            g0_blurred_image = masked_imaging_7x7.convolver.convolved_image_from_image_and_blurring_image(
                image=g0_image, blurring_image=g0_blurring_image
            )

            g1_image = g1.image_from_grid(grid=traced_grids_of_planes[1])
            g1_blurring_image = g1.image_from_grid(
                grid=traced_blurring_grids_of_planes[1]
            )

            g1_blurred_image = masked_imaging_7x7.convolver.convolved_image_from_image_and_blurring_image(
                image=g1_image, blurring_image=g1_blurring_image
            )

            assert fit.galaxy_model_image_dict[g0] == pytest.approx(
                g0_blurred_image, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[g1] == pytest.approx(
                g1_blurred_image, 1.0e-4
            )
            assert (fit.galaxy_model_image_dict[g2].in_1d == np.zeros(9)).all()

            assert fit.model_image.in_2d == pytest.approx(
                fit.galaxy_model_image_dict[g0].in_2d
                + fit.galaxy_model_image_dict[g1].in_2d,
                1.0e-4,
            )

        def test___all_lens_fit_quantities__including_hyper_methods(
            self, masked_imaging_7x7
        ):

            hyper_image_sky = al.hyper_data.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            image = hyper_image_sky.hyper_image_from_image(
                image=masked_imaging_7x7.image
            )

            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                mass_profile=al.mp.SphericalIsothermal(einstein_radius=1.0),
                hyper_galaxy=al.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image=np.ones(9),
                hyper_galaxy_image=np.ones(9),
                hyper_minimum_value=0.0,
            )
            g1 = al.Galaxy(
                redshift=1.0, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

            fit = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise_map_background = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            hyper_noise = tracer.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            hyper_noise_map = hyper_noise_map_background + hyper_noise

            assert hyper_noise_map.in_2d == pytest.approx(fit.noise_map.in_2d)

            model_image = tracer.blurred_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert model_image.in_2d == pytest.approx(fit.model_image.in_2d)

            residual_map = al.util.fit.residual_map_from(
                data=image, model_data=model_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(fit.chi_squared_map.in_2d)

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=hyper_noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)
            assert log_likelihood == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_image_properties(
            self, masked_imaging_7x7
        ):

            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                mass_profile=al.mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = al.Galaxy(
                redshift=1.0, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            blurred_images_of_planes = tracer.blurred_images_of_planes_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert blurred_images_of_planes[0].in_2d == pytest.approx(
                fit.model_images_of_planes[0].in_2d, 1.0e-4
            )

            assert blurred_images_of_planes[1].in_2d == pytest.approx(
                fit.model_images_of_planes[1].in_2d, 1.0e-4
            )

            unmasked_blurred_image = tracer.unmasked_blurred_image_from_grid_and_psf(
                grid=masked_imaging_7x7.grid, psf=masked_imaging_7x7.psf
            )

            assert (unmasked_blurred_image == fit.unmasked_blurred_image).all()

            unmasked_blurred_image_of_planes = tracer.unmasked_blurred_image_of_planes_from_grid_and_psf(
                grid=masked_imaging_7x7.grid, psf=masked_imaging_7x7.psf
            )

            assert (
                unmasked_blurred_image_of_planes[0]
                == fit.unmasked_blurred_image_of_planes[0]
            ).all()
            assert (
                unmasked_blurred_image_of_planes[1]
                == fit.unmasked_blurred_image_of_planes[1]
            ).all()

            unmasked_blurred_image_of_galaxies = tracer.unmasked_blurred_image_of_planes_and_galaxies_from_grid_and_psf(
                grid=masked_imaging_7x7.grid, psf=masked_imaging_7x7.psf
            )

            assert (
                unmasked_blurred_image_of_galaxies[0][0]
                == fit.unmasked_blurred_image_of_planes_and_galaxies[0][0]
            ).all()
            assert (
                unmasked_blurred_image_of_galaxies[1][0]
                == fit.unmasked_blurred_image_of_planes_and_galaxies[1][0]
            ).all()

    class TestCompareToManualInversionOnly:
        def test___all_lens_fit_quantities__no_hyper_methods(self, masked_imaging_7x7):

            # Ensures the inversion grid is used, as this would cause the test to fail.
            masked_imaging_7x7.grid[0, 0] = -100.0

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)

            g0 = al.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid_inversion, sparse_grid=None
            )
            inversion = inversions.InversionImagingMatrix.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image=masked_imaging_7x7.image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
            )

            assert inversion.mapped_reconstructed_image.in_2d == pytest.approx(
                fit.model_image.in_2d, 1.0e-4
            )

            residual_map = al.util.fit.residual_map_from(
                data=masked_imaging_7x7.image,
                model_data=inversion.mapped_reconstructed_image,
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d, 1.0e-4)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d, 1.0e-4
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(
                fit.chi_squared_map.in_2d, 1.0e-4
            )

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=masked_imaging_7x7.noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = al.util.fit.log_likelihood_with_regularization_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = al.util.fit.log_evidence_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

        def test___lens_fit_galaxy_model_image_dict__has_inversion_mapped_reconstructed_image(
            self, masked_imaging_7x7
        ):
            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)

            g0 = al.Galaxy(redshift=0.5)
            g1 = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid, sparse_grid=None
            )

            inversion = inversions.InversionImagingMatrix.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image=masked_imaging_7x7.image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
            )

            assert (fit.galaxy_model_image_dict[g0] == np.zeros(9)).all()

            assert fit.galaxy_model_image_dict[g1].in_2d == pytest.approx(
                inversion.mapped_reconstructed_image.in_2d, 1.0e-4
            )

            assert fit.model_image.in_2d == pytest.approx(
                fit.galaxy_model_image_dict[g1].in_2d, 1.0e-4
            )

        def test___all_lens_fit_quantities__include_hyper_methods(
            self, masked_imaging_7x7
        ):

            hyper_image_sky = al.hyper_data.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            image = hyper_image_sky.hyper_image_from_image(
                image=masked_imaging_7x7.image
            )

            hyper_noise_map_background = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)

            g0 = al.Galaxy(
                redshift=0.5,
                pixelization=pix,
                regularization=reg,
                hyper_galaxy=al.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image=np.ones(9),
                hyper_galaxy_image=np.ones(9),
                hyper_minimum_value=0.0,
            )

            tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

            fit = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise = tracer.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )
            hyper_noise_map = hyper_noise_map_background + hyper_noise

            assert hyper_noise_map.in_2d == pytest.approx(fit.noise_map.in_2d)

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid,
                settings=al.SettingsPixelization(use_border=False),
            )
            inversion = inversions.InversionImagingMatrix.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image=image,
                noise_map=hyper_noise_map,
                convolver=masked_imaging_7x7.convolver,
            )

            assert inversion.mapped_reconstructed_image.in_2d == pytest.approx(
                fit.model_image.in_2d, 1.0e-4
            )

            residual_map = al.util.fit.residual_map_from(
                data=image, model_data=inversion.mapped_reconstructed_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(fit.chi_squared_map.in_2d)

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=hyper_noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = al.util.fit.log_likelihood_with_regularization_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = al.util.fit.log_evidence_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_image_properties(
            self, masked_imaging_7x7
        ):

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)

            g0 = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid,
                settings=al.SettingsPixelization(use_border=False),
            )

            inversion = inversions.InversionImagingMatrix.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                image=masked_imaging_7x7.image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
            )

            assert (fit.model_images_of_planes[0].in_2d == np.zeros((7, 7))).all()
            assert inversion.mapped_reconstructed_image.in_2d == pytest.approx(
                fit.model_images_of_planes[1].in_2d, 1.0e-4
            )

        def test___stochastic_mode_gives_different_log_likelihoods(
            self, masked_imaging_7x7
        ):

            # Ensures the inversion grid is used, as this would cause the test to fail.
            masked_imaging_7x7.grid[0, 0] = -100.0

            pix = al.pix.VoronoiBrightnessImage(pixels=9)
            reg = al.reg.Constant(coefficient=1.0)

            g0 = al.Galaxy(
                redshift=0.5,
                pixelization=pix,
                regularization=reg,
                hyper_model_image=al.Array.ones(shape_2d=(3, 3)),
                hyper_galaxy_image=al.Array.ones(shape_2d=(3, 3)),
            )

            tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

            fit_0 = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                settings_pixelization=al.SettingsPixelization(is_stochastic=False),
            )
            fit_1 = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                settings_pixelization=al.SettingsPixelization(is_stochastic=False),
            )

            assert fit_0.log_evidence == fit_1.log_evidence

            fit_0 = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                settings_pixelization=al.SettingsPixelization(is_stochastic=True),
            )
            fit_1 = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                settings_pixelization=al.SettingsPixelization(is_stochastic=True),
            )

            assert fit_0.log_evidence != fit_1.log_evidence

    class TestCompareToManualProfilesAndInversion:
        def test___all_lens_fit_quantities__no_hyper_methods(self, masked_imaging_7x7):
            galaxy_light = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)
            galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            blurred_image = tracer.blurred_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert blurred_image.in_2d == pytest.approx(fit.blurred_image.in_2d)

            profile_subtracted_image = masked_imaging_7x7.image - blurred_image

            assert profile_subtracted_image.in_2d == pytest.approx(
                fit.profile_subtracted_image.in_2d
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid,
                settings=al.SettingsPixelization(use_border=False),
            )

            inversion = inversions.InversionImagingMatrix.from_data_mapper_and_regularization(
                image=profile_subtracted_image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
                mapper=mapper,
                regularization=reg,
            )

            model_image = blurred_image + inversion.mapped_reconstructed_image

            assert model_image.in_2d == pytest.approx(fit.model_image.in_2d)

            residual_map = al.util.fit.residual_map_from(
                data=masked_imaging_7x7.image, model_data=model_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=masked_imaging_7x7.noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(fit.chi_squared_map.in_2d)

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=masked_imaging_7x7.noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = al.util.fit.log_likelihood_with_regularization_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = al.util.fit.log_evidence_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

        def test___lens_fit_galaxy_model_image_dict__has_blurred_images_and_inversion_mapped_reconstructed_image(
            self, masked_imaging_7x7
        ):

            g0 = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )
            g1 = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=2.0)
            )
            g2 = al.Galaxy(redshift=0.5)

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)
            galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2, galaxy_pix])

            masked_imaging_7x7.image[0] = 3.0

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            traced_grids = tracer.traced_grids_of_planes_from_grid(
                grid=masked_imaging_7x7.grid
            )
            traced_blurring_grids = tracer.traced_grids_of_planes_from_grid(
                grid=masked_imaging_7x7.blurring_grid
            )

            g0_image = g0.image_from_grid(grid=traced_grids[0])
            g0_blurring_image = g0.image_from_grid(grid=traced_blurring_grids[0])

            g0_blurred_image = masked_imaging_7x7.convolver.convolved_image_from_image_and_blurring_image(
                image=g0_image, blurring_image=g0_blurring_image
            )

            g1_image = g1.image_from_grid(grid=traced_grids[1])
            g1_blurring_image = g1.image_from_grid(grid=traced_blurring_grids[1])

            g1_blurred_image = masked_imaging_7x7.convolver.convolved_image_from_image_and_blurring_image(
                image=g1_image, blurring_image=g1_blurring_image
            )

            blurred_image = g0_blurred_image + g1_blurred_image

            profile_subtracted_image = masked_imaging_7x7.image - blurred_image

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid,
                settings=al.SettingsPixelization(use_border=False),
            )

            inversion = inversions.InversionImagingMatrix.from_data_mapper_and_regularization(
                image=profile_subtracted_image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
                mapper=mapper,
                regularization=reg,
            )

            assert (fit.galaxy_model_image_dict[g2] == np.zeros(9)).all()

            assert fit.galaxy_model_image_dict[g0].in_2d == pytest.approx(
                g0_blurred_image.in_2d, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[g1].in_2d == pytest.approx(
                g1_blurred_image.in_2d, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[galaxy_pix].in_2d == pytest.approx(
                inversion.mapped_reconstructed_image.in_2d, 1.0e-4
            )

            assert fit.model_image.in_2d == pytest.approx(
                fit.galaxy_model_image_dict[g0].in_2d
                + fit.galaxy_model_image_dict[g1].in_2d
                + inversion.mapped_reconstructed_image.in_2d,
                1.0e-4,
            )

        def test___all_lens_fit_quantities__include_hyper_methods(
            self, masked_imaging_7x7
        ):

            hyper_image_sky = al.hyper_data.HyperImageSky(sky_scale=1.0)

            hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            image = hyper_image_sky.hyper_image_from_image(
                image=masked_imaging_7x7.image
            )

            hyper_noise_map_background = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )

            galaxy_light = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                hyper_galaxy=al.HyperGalaxy(
                    contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
                ),
                hyper_model_image=al.Array.ones(shape_2d=(3, 3)),
                hyper_galaxy_image=al.Array.ones(shape_2d=(3, 3)),
                hyper_minimum_value=0.0,
            )

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)
            galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])

            fit = al.FitImaging(
                masked_imaging=masked_imaging_7x7,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            hyper_noise = tracer.hyper_noise_map_from_noise_map(
                noise_map=masked_imaging_7x7.noise_map
            )
            hyper_noise_map = hyper_noise_map_background + hyper_noise

            assert hyper_noise_map.in_2d == pytest.approx(fit.noise_map.in_2d, 1.0e-4)

            blurred_image = tracer.blurred_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            assert blurred_image.in_2d == pytest.approx(fit.blurred_image.in_2d)

            profile_subtracted_image = image - blurred_image

            assert profile_subtracted_image.in_2d == pytest.approx(
                fit.profile_subtracted_image.in_2d
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid,
                settings=al.SettingsPixelization(use_border=False),
            )

            inversion = inversions.InversionImagingMatrix.from_data_mapper_and_regularization(
                image=profile_subtracted_image,
                noise_map=hyper_noise_map,
                convolver=masked_imaging_7x7.convolver,
                mapper=mapper,
                regularization=reg,
            )

            model_image = blurred_image + inversion.mapped_reconstructed_image

            assert model_image.in_2d == pytest.approx(fit.model_image.in_2d, 1.0e-4)

            residual_map = al.util.fit.residual_map_from(
                data=image, model_data=model_image
            )

            assert residual_map.in_2d == pytest.approx(fit.residual_map.in_2d, 1.0e-4)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert normalized_residual_map.in_2d == pytest.approx(
                fit.normalized_residual_map.in_2d, 1.0e-4
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=hyper_noise_map
            )

            assert chi_squared_map.in_2d == pytest.approx(
                fit.chi_squared_map.in_2d, 1.0e-4
            )

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=hyper_noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = al.util.fit.log_likelihood_with_regularization_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = al.util.fit.log_evidence_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

        def test___blurred_and_model_images_of_planes_and_unmasked_blurred_image_properties(
            self, masked_imaging_7x7
        ):
            galaxy_light = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)
            galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])

            fit = al.FitImaging(masked_imaging=masked_imaging_7x7, tracer=tracer)

            blurred_image = tracer.blurred_image_from_grid_and_convolver(
                grid=masked_imaging_7x7.grid,
                convolver=masked_imaging_7x7.convolver,
                blurring_grid=masked_imaging_7x7.blurring_grid,
            )

            profile_subtracted_image = masked_imaging_7x7.image - blurred_image

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_imaging_7x7.grid,
                settings=al.SettingsPixelization(use_border=False),
            )

            inversion = inversions.InversionImagingMatrix.from_data_mapper_and_regularization(
                image=profile_subtracted_image,
                noise_map=masked_imaging_7x7.noise_map,
                convolver=masked_imaging_7x7.convolver,
                mapper=mapper,
                regularization=reg,
            )

            assert blurred_image.in_2d == pytest.approx(
                fit.model_images_of_planes[0].in_2d, 1.0e-4
            )
            assert inversion.mapped_reconstructed_image.in_2d == pytest.approx(
                fit.model_images_of_planes[1].in_2d, 1.0e-4
            )


class TestFitInterferometer:
    class TestFitProperties:
        def test__total_inversions(self, masked_interferometer_7):
            g0 = al.Galaxy(redshift=0.5)

            g1 = al.Galaxy(redshift=1.0)

            g2 = al.Galaxy(redshift=2.0)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7, tracer=tracer
            )

            assert fit.total_inversions == 0

            g2 = al.Galaxy(
                redshift=2.0,
                pixelization=al.pix.Rectangular(),
                regularization=al.reg.Constant(),
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7, tracer=tracer
            )

            assert fit.total_inversions == 1

            g0 = al.Galaxy(
                redshift=0.5,
                pixelization=al.pix.Rectangular(),
                regularization=al.reg.Constant(),
            )

            g1 = al.Galaxy(
                redshift=1.0,
                pixelization=al.pix.Rectangular(),
                regularization=al.reg.Constant(),
            )

            g2 = al.Galaxy(
                redshift=2.0,
                pixelization=al.pix.Rectangular(),
                regularization=al.reg.Constant(),
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7, tracer=tracer
            )

            assert fit.total_inversions == 3

    class TestLikelihood:
        def test__1x2_image__1x2_visibilities__simple_fourier_transform(self):
            # The image plane image generated by the galaxy is [1.0, 1.0]

            # Thus the chi squared is 4.0**2.0 + 3.0**2.0 = 25.0

            uv_wavelengths = np.array([[0.0, 0.0]])

            interferometer = al.Interferometer(
                visibilities=al.Visibilities.full(fill_value=5.0, shape_1d=(1,)),
                noise_map=al.Visibilities.ones(shape_1d=(1,)),
                uv_wavelengths=uv_wavelengths,
            )

            interferometer.visibilities[0, 1] = 4.0

            visibilities_mask = np.full(fill_value=False, shape=(1, 2))

            real_space_mask = al.Mask.manual(
                mask=np.array(
                    [
                        [True, True, True, True],
                        [True, False, False, True],
                        [True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
            )

            masked_interferometer = al.MaskedInterferometer(
                interferometer=interferometer,
                visibilities_mask=visibilities_mask,
                real_space_mask=real_space_mask,
                settings=al.SettingsMaskedInterferometer(
                    sub_size=1, transformer_class=al.TransformerDFT
                ),
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = al.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = al.Tracer.from_galaxies(galaxies=[g0])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer, tracer=tracer
            )

            assert (fit.visibilities_mask == np.array([False, False])).all()

            assert (fit.visibilities.in_1d == np.array([[5.0, 4.0]])).all()
            assert (fit.noise_map.in_1d == np.array([[1.0, 1.0]])).all()
            assert (fit.model_visibilities.in_1d == np.array([2.0, 0.0])).all()
            assert (fit.residual_map.in_1d == np.array([3.0, 4.0])).all()
            assert (fit.normalized_residual_map.in_1d == np.array([3.0, 4.0])).all()
            assert (fit.chi_squared_map.in_1d == np.array([9.0, 16.0])).all()

            assert fit.chi_squared == 25.0
            assert fit.reduced_chi_squared == 25.0 / 2.0
            assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
            assert fit.log_likelihood == -0.5 * (
                25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0)
            )

        def test__3x2_image__2x2_visibilities__use_transformer_for_likelihood(self):

            uv_wavelengths = np.array([[1.0, 0.0], [1.0, 1.0], [2.0, 2.0]])

            interferometer = al.Interferometer(
                visibilities=al.Visibilities.full(fill_value=5.0, shape_1d=(3,)),
                noise_map=al.Visibilities.full(fill_value=2.0, shape_1d=(3,)),
                uv_wavelengths=uv_wavelengths,
            )

            real_space_mask = al.Mask.unmasked(shape_2d=(1, 3), pixel_scales=1.0)

            transformer = al.TransformerDFT(
                uv_wavelengths=uv_wavelengths, real_space_mask=real_space_mask
            )

            visibilities_mask = np.full(fill_value=False, shape=(3, 2))

            real_space_mask = al.Mask.manual(
                mask=np.array(
                    [
                        [True, True, True, True, True],
                        [True, False, False, False, True],
                        [True, True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
            )

            masked_interferometer = al.MaskedInterferometer(
                interferometer=interferometer,
                visibilities_mask=visibilities_mask,
                real_space_mask=real_space_mask,
                settings=al.SettingsMaskedInterferometer(
                    sub_size=1, transformer_class=al.TransformerDFT
                ),
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=0.001)
            )

            image = g0.image_from_grid(grid=masked_interferometer.grid)

            model_visibilities_manual = transformer.visibilities_from_image(image=image)

            tracer = al.Tracer.from_galaxies(galaxies=[g0])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer, tracer=tracer
            )

            assert (fit.visibilities_mask == np.array([False, False])).all()

            assert (
                fit.visibilities.in_1d
                == np.array([[[5.0, 5.0], [5.0, 5.0], [5.0, 5.0]]])
            ).all()
            assert (
                fit.noise_map.in_1d == np.array([[[2.0, 2.0], [2.0, 2.0], [2.0, 2.0]]])
            ).all()

            assert fit.model_visibilities.in_1d == pytest.approx(
                model_visibilities_manual, 1.0e-4
            )

            # moddel visibilities are all [1.94805, 0.0]

            assert fit.residual_map.in_1d == pytest.approx(
                np.array([[3.0519, 5.0], [3.0519, 5.0], [3.0519, 5.0]]), 1.0e-4
            )

            assert fit.normalized_residual_map.in_1d == pytest.approx(
                np.array(
                    [
                        [3.0519 / 2.0, 5.0 / 2.0],
                        [3.0519 / 2.0, 5.0 / 2.0],
                        [3.0519 / 2.0, 5.0 / 2.0],
                    ]
                ),
                1.0e-4,
            )

            assert fit.chi_squared_map.in_1d == pytest.approx(
                np.array(
                    [
                        [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                        [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                        [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                    ]
                ),
                1.0e-4,
            )

            assert fit.chi_squared == pytest.approx(25.73579, 1.0e-4)
            assert fit.reduced_chi_squared == pytest.approx(25.73579 / 6.0)
            assert fit.noise_normalization == pytest.approx(
                (6.0 * np.log(2 * np.pi * 2.0 ** 2.0)), 1.0e-4
            )
            assert fit.log_likelihood == pytest.approx(
                -0.5 * (25.73579 + 6.0 * np.log(2 * np.pi * 2.0 ** 2.0)), 1.0e-4
            )

        def test__hyper_background_changes_background_sky__reflected_in_likelihood(
            self
        ):

            uv_wavelengths = np.array([[1.0, 0.0], [1.0, 1.0], [2.0, 2.0]])

            interferometer = al.Interferometer(
                visibilities=al.Visibilities.full(fill_value=5.0, shape_1d=(3,)),
                noise_map=al.Visibilities.full(fill_value=2.0, shape_1d=(3,)),
                uv_wavelengths=uv_wavelengths,
            )

            visibilities_mask = np.full(fill_value=False, shape=(1, 2))

            real_space_mask = al.Mask.manual(
                mask=np.array(
                    [
                        [True, True, True, True, True],
                        [True, False, False, False, True],
                        [True, True, True, True, True],
                    ]
                ),
                pixel_scales=1.0,
            )

            masked_interferometer = al.MaskedInterferometer(
                interferometer=interferometer,
                visibilities_mask=visibilities_mask,
                real_space_mask=real_space_mask,
                settings=al.SettingsMaskedInterferometer(sub_size=1),
            )

            # Setup as a ray trace instance, using a light profile for the lens

            g0 = al.Galaxy(
                redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2)
            )
            tracer = al.Tracer.from_galaxies(galaxies=[g0])

            hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer,
                tracer=tracer,
                hyper_background_noise=hyper_background_noise,
            )

            assert (
                fit.visibilities.in_1d == np.full(fill_value=5.0, shape=(3, 2))
            ).all()

            assert (fit.noise_map.in_1d == np.full(fill_value=3.0, shape=(3, 2))).all()

    class TestCompareToManualProfilesOnly:
        def test___all_lens_fit_quantities__no_hyper_methods(
            self, masked_interferometer_7
        ):
            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                mass_profile=al.mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = al.Galaxy(
                redshift=1.0, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7, tracer=tracer
            )

            assert masked_interferometer_7.noise_map == pytest.approx(fit.noise_map)

            model_visibilities = tracer.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            assert model_visibilities == pytest.approx(fit.model_visibilities, 1e-4)

            residual_map = al.util.fit.residual_map_from(
                data=masked_interferometer_7.visibilities, model_data=model_visibilities
            )

            assert residual_map == pytest.approx(fit.residual_map, 1e-4)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert normalized_residual_map == pytest.approx(
                fit.normalized_residual_map, 1e-4
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert chi_squared_map == pytest.approx(fit.chi_squared_map, 1e-4)

            chi_squared = al.util.fit.chi_squared_from(
                chi_squared_map=fit.chi_squared_map
            )

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=masked_interferometer_7.noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)
            assert log_likelihood == fit.figure_of_merit

        def test___lens_fit_galaxy_model_image_dict__corresponds_to_profile_galaxy_images(
            self, masked_interferometer_7_grid
        ):
            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                mass_profile=al.mp.SphericalIsothermal(einstein_radius=1.0),
            )
            g1 = al.Galaxy(
                redshift=1.0, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )
            g2 = al.Galaxy(redshift=1.0)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7_grid, tracer=tracer
            )

            traced_grids_of_planes = tracer.traced_grids_of_planes_from_grid(
                grid=masked_interferometer_7_grid.grid
            )

            g0_image = g0.image_from_grid(grid=traced_grids_of_planes[0])

            g1_image = g1.image_from_grid(grid=traced_grids_of_planes[1])

            assert fit.galaxy_model_image_dict[g0].in_1d == pytest.approx(
                g0_image, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[g1].in_1d == pytest.approx(
                g1_image, 1.0e-4
            )

        def test___lens_fit_galaxy_visibilities_dict__corresponds_to_galaxy_visibilities(
            self, masked_interferometer_7_grid
        ):
            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                mass_profile=al.mp.SphericalIsothermal(einstein_radius=1.0),
            )
            g1 = al.Galaxy(
                redshift=1.0, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )
            g2 = al.Galaxy(redshift=1.0)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7_grid, tracer=tracer
            )

            traced_grids_of_planes = tracer.traced_grids_of_planes_from_grid(
                grid=masked_interferometer_7_grid.grid
            )

            g0_profile_visibilities = g0.profile_visibilities_from_grid_and_transformer(
                grid=traced_grids_of_planes[0],
                transformer=masked_interferometer_7_grid.transformer,
            )

            g1_profile_visibilities = g1.profile_visibilities_from_grid_and_transformer(
                grid=traced_grids_of_planes[1],
                transformer=masked_interferometer_7_grid.transformer,
            )

            assert fit.galaxy_model_visibilities_dict[g0].in_1d == pytest.approx(
                g0_profile_visibilities, 1.0e-4
            )
            assert fit.galaxy_model_visibilities_dict[g1].in_1d == pytest.approx(
                g1_profile_visibilities, 1.0e-4
            )
            assert (
                fit.galaxy_model_visibilities_dict[g2].in_1d == np.zeros((7, 2))
            ).all()

            assert fit.model_visibilities.in_1d == pytest.approx(
                fit.galaxy_model_visibilities_dict[g0].in_1d
                + fit.galaxy_model_visibilities_dict[g1].in_1d,
                1.0e-4,
            )

        def test___all_lens_fit_quantities__hyper_background_noise(
            self, masked_interferometer_7
        ):

            hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            hyper_noise_map = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            g0 = al.Galaxy(
                redshift=0.5,
                light_profile=al.lp.EllipticalSersic(intensity=1.0),
                mass_profile=al.mp.SphericalIsothermal(einstein_radius=1.0),
            )

            g1 = al.Galaxy(
                redshift=1.0, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                tracer=tracer,
                hyper_background_noise=hyper_background_noise,
            )

            assert hyper_noise_map.in_1d == pytest.approx(fit.noise_map.in_1d)

    class TestCompareToManualInversionOnly:
        def test___all_lens_fit_quantities__no_hyper_methods(
            self, masked_interferometer_7
        ):

            # Ensures the inversion grid is used, as this would cause the test to fail.
            masked_interferometer_7.grid[0, 0] = -100.0

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=0.01)

            g0 = al.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7, tracer=tracer
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid_inversion, sparse_grid=None
            )

            inversion = inversions.InversionInterferometerMatrix.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                visibilities=masked_interferometer_7.visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
            )

            assert inversion.mapped_reconstructed_visibilities == pytest.approx(
                fit.model_visibilities, 1.0e-4
            )

            residual_map = al.util.fit.residual_map_from(
                data=masked_interferometer_7.visibilities,
                model_data=inversion.mapped_reconstructed_visibilities,
            )

            assert residual_map.in_1d == pytest.approx(fit.residual_map.in_1d, 1.0e-4)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert normalized_residual_map.in_1d == pytest.approx(
                fit.normalized_residual_map.in_1d, 1.0e-4
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert chi_squared_map.in_1d == pytest.approx(
                fit.chi_squared_map.in_1d, 1.0e-4
            )

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=masked_interferometer_7.noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = al.util.fit.log_likelihood_with_regularization_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = al.util.fit.log_evidence_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

            mapped_reconstructed_image = al.util.inversion.mapped_reconstructed_data_from(
                mapping_matrix=fit.inversion.mapper.mapping_matrix,
                reconstruction=fit.inversion.reconstruction,
            )

            assert (
                fit.inversion.mapped_reconstructed_image.in_1d
                == mapped_reconstructed_image
            ).all()

        def test___lens_fit_galaxy_model_image_dict__images_and_inversion_mapped_reconstructed_image(
            self, masked_interferometer_7
        ):
            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)

            g0 = al.Galaxy(redshift=0.5)
            g1 = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7, tracer=tracer
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid, sparse_grid=None
            )

            inversion = inversions.InversionInterferometerMatrix.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                visibilities=masked_interferometer_7.visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
            )

            assert (fit.galaxy_model_image_dict[g0].in_2d == np.zeros((7, 7))).all()

            assert fit.galaxy_model_image_dict[g1].in_1d == pytest.approx(
                inversion.mapped_reconstructed_image.in_1d, 1.0e-4
            )

        def test___lens_fit_galaxy_model_visibilities_dict__has_inversion_mapped_reconstructed_visibilities(
            self, masked_interferometer_7
        ):
            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)

            g0 = al.Galaxy(redshift=0.5)
            g1 = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7, tracer=tracer
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid, sparse_grid=None
            )

            inversion = inversions.InversionInterferometerMatrix.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                visibilities=masked_interferometer_7.visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
            )

            assert (fit.galaxy_model_visibilities_dict[g0] == np.zeros((7, 2))).all()

            assert fit.galaxy_model_visibilities_dict[g1].in_1d == pytest.approx(
                inversion.mapped_reconstructed_visibilities.in_1d, 1.0e-4
            )

            assert fit.model_visibilities.in_1d == pytest.approx(
                fit.galaxy_model_visibilities_dict[g1].in_1d, 1.0e-4
            )

        def test___all_lens_fit_quantities__hyper_background_noise(
            self, masked_interferometer_7
        ):

            hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            hyper_noise_map = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=0.01)

            g0 = al.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                tracer=tracer,
                hyper_background_noise=hyper_background_noise,
            )

            assert hyper_noise_map.in_1d == pytest.approx(
                fit.inversion.noise_map, 1.0e-4
            )

            assert hyper_noise_map.in_1d == pytest.approx(fit.noise_map.in_1d)

        def test___stochastic_mode_gives_different_log_likelihoods(
            self, masked_interferometer_7
        ):

            # Ensures the inversion grid is used, as this would cause the test to fail.
            masked_interferometer_7.grid[0, 0] = -100.0

            pix = al.pix.VoronoiBrightnessImage(pixels=9)
            reg = al.reg.Constant(coefficient=1.0)

            g0 = al.Galaxy(
                redshift=0.5,
                pixelization=pix,
                regularization=reg,
                hyper_model_image=al.Array.ones(shape_2d=(3, 3)),
                hyper_galaxy_image=al.Array.ones(shape_2d=(3, 3)),
            )

            tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

            fit_0 = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                tracer=tracer,
                settings_pixelization=al.SettingsPixelization(is_stochastic=False),
            )
            fit_1 = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                tracer=tracer,
                settings_pixelization=al.SettingsPixelization(is_stochastic=False),
            )

            assert fit_0.log_evidence == fit_1.log_evidence

            fit_0 = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                tracer=tracer,
                settings_pixelization=al.SettingsPixelization(is_stochastic=True),
            )
            fit_1 = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                tracer=tracer,
                settings_pixelization=al.SettingsPixelization(is_stochastic=True),
            )

            assert fit_0.log_evidence != fit_1.log_evidence

        def test___all_lens_fit_quantities__linear_operator(
            self, masked_interferometer_7
        ):

            # Ensures the inversion grid is used, as this would cause the test to fail.
            masked_interferometer_7.grid[0, 0] = -100.0

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=0.01)

            g0 = al.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                tracer=tracer,
                settings_inversion=al.SettingsInversion(use_linear_operators=True),
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid_inversion, sparse_grid=None
            )

            inversion = inversions.InversionInterferometerLinearOperator.from_data_mapper_and_regularization(
                mapper=mapper,
                regularization=reg,
                visibilities=masked_interferometer_7.visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
                settings=al.SettingsInversion(use_linear_operators=True),
            )

            assert inversion.mapped_reconstructed_visibilities == pytest.approx(
                fit.model_visibilities, 1.0e-4
            )

            residual_map = al.util.fit.residual_map_from(
                data=masked_interferometer_7.visibilities,
                model_data=inversion.mapped_reconstructed_visibilities,
            )

            assert residual_map.in_1d == pytest.approx(fit.residual_map.in_1d, 1.0e-4)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert normalized_residual_map.in_1d == pytest.approx(
                fit.normalized_residual_map.in_1d, 1.0e-4
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert chi_squared_map.in_1d == pytest.approx(
                fit.chi_squared_map.in_1d, 1.0e-4
            )

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=masked_interferometer_7.noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = al.util.fit.log_likelihood_with_regularization_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = al.util.fit.log_evidence_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

            mapped_reconstructed_image = al.util.inversion.mapped_reconstructed_data_from(
                mapping_matrix=fit.inversion.mapper.mapping_matrix,
                reconstruction=fit.inversion.reconstruction,
            )

            assert (
                fit.inversion.mapped_reconstructed_image.in_1d
                == mapped_reconstructed_image
            ).all()

    class TestCompareToManualProfilesAndInversion:
        def test___all_lens_fit_quantities__no_hyper_methods(
            self, masked_interferometer_7
        ):
            galaxy_light = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)
            galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7, tracer=tracer
            )

            profile_visibilities = tracer.profile_visibilities_from_grid_and_transformer(
                grid=masked_interferometer_7.grid,
                transformer=masked_interferometer_7.transformer,
            )

            assert profile_visibilities.in_1d == pytest.approx(
                fit.profile_visibilities.in_1d
            )

            profile_subtracted_visibilities = (
                masked_interferometer_7.visibilities - profile_visibilities
            )

            assert profile_subtracted_visibilities.in_1d == pytest.approx(
                fit.profile_subtracted_visibilities.in_1d
            )

            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7.grid,
                settings=al.SettingsPixelization(use_border=False),
            )

            inversion = inversions.InversionInterferometerMatrix.from_data_mapper_and_regularization(
                visibilities=profile_subtracted_visibilities,
                noise_map=masked_interferometer_7.noise_map,
                transformer=masked_interferometer_7.transformer,
                mapper=mapper,
                regularization=reg,
            )

            model_visibilities = (
                profile_visibilities + inversion.mapped_reconstructed_visibilities
            )

            assert model_visibilities.in_1d == pytest.approx(
                fit.model_visibilities.in_1d
            )

            residual_map = al.util.fit.residual_map_from(
                data=masked_interferometer_7.visibilities, model_data=model_visibilities
            )

            assert residual_map.in_1d == pytest.approx(fit.residual_map.in_1d)

            normalized_residual_map = al.util.fit.normalized_residual_map_from(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert normalized_residual_map.in_1d == pytest.approx(
                fit.normalized_residual_map.in_1d
            )

            chi_squared_map = al.util.fit.chi_squared_map_from(
                residual_map=residual_map, noise_map=masked_interferometer_7.noise_map
            )

            assert chi_squared_map.in_1d == pytest.approx(fit.chi_squared_map.in_1d)

            chi_squared = al.util.fit.chi_squared_from(chi_squared_map=chi_squared_map)

            noise_normalization = al.util.fit.noise_normalization_from(
                noise_map=masked_interferometer_7.noise_map
            )

            log_likelihood = al.util.fit.log_likelihood_from(
                chi_squared=chi_squared, noise_normalization=noise_normalization
            )

            assert log_likelihood == pytest.approx(fit.log_likelihood, 1e-4)

            log_likelihood_with_regularization = al.util.fit.log_likelihood_with_regularization_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                noise_normalization=noise_normalization,
            )

            assert log_likelihood_with_regularization == pytest.approx(
                fit.log_likelihood_with_regularization, 1e-4
            )

            log_evidence = al.util.fit.log_evidence_from(
                chi_squared=chi_squared,
                regularization_term=inversion.regularization_term,
                log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
                log_regularization_term=inversion.log_det_regularization_matrix_term,
                noise_normalization=noise_normalization,
            )

            assert log_evidence == fit.log_evidence
            assert log_evidence == fit.figure_of_merit

            mapped_reconstructed_image = al.util.inversion.mapped_reconstructed_data_from(
                mapping_matrix=fit.inversion.mapper.mapping_matrix,
                reconstruction=fit.inversion.reconstruction,
            )

            assert (
                fit.inversion.mapped_reconstructed_image.in_1d
                == mapped_reconstructed_image
            ).all()

        def test___lens_fit_galaxy_model_visibilities_dict__has_image_and_inversion_mapped_reconstructed_image(
            self, masked_interferometer_7_grid
        ):

            g0 = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )
            g1 = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=2.0)
            )
            g2 = al.Galaxy(redshift=0.5)

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)
            galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2, galaxy_pix])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7_grid, tracer=tracer
            )

            traced_grids = tracer.traced_grids_of_planes_from_grid(
                grid=masked_interferometer_7_grid.grid
            )

            g0_visibilities = g0.profile_visibilities_from_grid_and_transformer(
                grid=traced_grids[0],
                transformer=masked_interferometer_7_grid.transformer,
            )

            g1_visibilities = g1.profile_visibilities_from_grid_and_transformer(
                grid=traced_grids[1],
                transformer=masked_interferometer_7_grid.transformer,
            )

            profile_visibilities = g0_visibilities + g1_visibilities

            profile_subtracted_visibilities = (
                masked_interferometer_7_grid.visibilities - profile_visibilities
            )
            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7_grid.grid,
                settings=al.SettingsPixelization(use_border=False),
            )

            inversion = inversions.InversionInterferometerMatrix.from_data_mapper_and_regularization(
                visibilities=profile_subtracted_visibilities,
                noise_map=masked_interferometer_7_grid.noise_map,
                transformer=masked_interferometer_7_grid.transformer,
                mapper=mapper,
                regularization=reg,
            )

            g0_image = g0.image_from_grid(grid=traced_grids[0])

            g1_image = g1.image_from_grid(grid=traced_grids[1])

            assert (fit.galaxy_model_image_dict[g2].in_2d == np.zeros((7, 7))).all()

            assert fit.galaxy_model_image_dict[g0].in_1d == pytest.approx(
                g0_image.in_1d, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[g1].in_1d == pytest.approx(
                g1_image.in_1d, 1.0e-4
            )
            assert fit.galaxy_model_image_dict[galaxy_pix].in_1d == pytest.approx(
                inversion.mapped_reconstructed_image.in_1d, 1.0e-4
            )

        def test___lens_fit_galaxy_model_visibilities_dict__has_profile_visibilitiess_and_inversion_mapped_reconstructed_visibilities(
            self, masked_interferometer_7_grid
        ):

            g0 = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )
            g1 = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=2.0)
            )
            g2 = al.Galaxy(redshift=0.5)

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)
            galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2, galaxy_pix])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7_grid, tracer=tracer
            )

            traced_grids = tracer.traced_grids_of_planes_from_grid(
                grid=masked_interferometer_7_grid.grid
            )

            g0_visibilities = g0.profile_visibilities_from_grid_and_transformer(
                grid=traced_grids[0],
                transformer=masked_interferometer_7_grid.transformer,
            )

            g1_visibilities = g1.profile_visibilities_from_grid_and_transformer(
                grid=traced_grids[1],
                transformer=masked_interferometer_7_grid.transformer,
            )

            profile_visibilities = g0_visibilities + g1_visibilities

            profile_subtracted_visibilities = (
                masked_interferometer_7_grid.visibilities - profile_visibilities
            )
            mapper = pix.mapper_from_grid_and_sparse_grid(
                grid=masked_interferometer_7_grid.grid,
                settings=al.SettingsPixelization(use_border=False),
            )

            inversion = inversions.InversionInterferometerMatrix.from_data_mapper_and_regularization(
                visibilities=profile_subtracted_visibilities,
                noise_map=masked_interferometer_7_grid.noise_map,
                transformer=masked_interferometer_7_grid.transformer,
                mapper=mapper,
                regularization=reg,
            )

            assert (fit.galaxy_model_visibilities_dict[g2] == np.zeros((7, 2))).all()

            assert fit.galaxy_model_visibilities_dict[g0].in_1d == pytest.approx(
                g0_visibilities.in_1d, 1.0e-4
            )
            assert fit.galaxy_model_visibilities_dict[g1].in_1d == pytest.approx(
                g1_visibilities.in_1d, 1.0e-4
            )
            assert fit.galaxy_model_visibilities_dict[
                galaxy_pix
            ].in_1d == pytest.approx(
                inversion.mapped_reconstructed_visibilities.in_1d, 1.0e-4
            )

            assert fit.model_visibilities.in_1d == pytest.approx(
                fit.galaxy_model_visibilities_dict[g0].in_1d
                + fit.galaxy_model_visibilities_dict[g1].in_1d
                + inversion.mapped_reconstructed_visibilities.in_1d,
                1.0e-4,
            )

        def test___all_lens_fit_quantities__hyper_background_noise(
            self, masked_interferometer_7
        ):

            hyper_background_noise = al.hyper_data.HyperBackgroundNoise(noise_scale=1.0)

            hyper_noise_map = hyper_background_noise.hyper_noise_map_from_noise_map(
                noise_map=masked_interferometer_7.noise_map
            )

            galaxy_light = al.Galaxy(
                redshift=0.5, light_profile=al.lp.EllipticalSersic(intensity=1.0)
            )

            pix = al.pix.Rectangular(shape=(3, 3))
            reg = al.reg.Constant(coefficient=1.0)
            galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

            tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])

            fit = al.FitInterferometer(
                masked_interferometer=masked_interferometer_7,
                tracer=tracer,
                hyper_background_noise=hyper_background_noise,
            )

            assert hyper_noise_map.in_1d == pytest.approx(
                fit.inversion.noise_map, 1.0e-4
            )

            assert hyper_noise_map.in_1d == pytest.approx(fit.noise_map.in_1d)



from test_autolens.simulators.interferometer import simulators

"""
Welcome to the PyAutoLens dataset generator. Here, we'll make the datasets that we use to test and profile
PyAutoLens. This consists of the following sets of visibilitiess:

- Source-only visibilities, where the lens mass is an SIE and the source light is a smooth Exponential.
- Source-only visibilities, where the lens mass is an SIE and source light a cuspy Sersic (sersic_index=3).
- The same smooth source visibilities above, but with an offset lens / source centre such that the 
  lens / source galaxy visibilitiess are not as the centre of the visibilities.
- Lens + source visibilities, where the lens light is a Dev Vaucouleurs, mass is an SIE and the source light is a smooth
  Exponential.
- Lens + source visibilities, where the lens light is a Dev Vaucouleurs, mass is an SIE and source light a cuspy Sersic
  (sersic_index=3).

Each visibilities is generated for one instrument.
"""

# To simulate each lens, we pass it an instrument and call its simulate function.

instruments = ["sma"]

for instrument in instruments:

    simulators.simulate__lens_sie__source_smooth(instrument=instrument)
    simulators.simulate__lens_sie__source_cuspy(instrument=instrument)
    simulators.simulate__lens_sie__source_smooth__offset_centre(instrument=instrument)
    simulators.simulate__lens_light__source_smooth(instrument=instrument)
    simulators.simulate__lens_light__source_cuspy(instrument=instrument)



import autoarray as aa
from autolens.fit import fit_positions
from autogalaxy.pipeline.phase import dataset
from autolens import exc
import numpy as np

import copy


class PhaseDataset(dataset.PhaseDataset):
    def modify_dataset(self, dataset, results):

        # TODO : There is a very weird error no cosma for this line we don't yet undersatand. This try / except fixes it.

        try:

            positions_copy = copy.copy(dataset.positions)

            positions = self.updated_positions_from_positions_and_results(
                positions=positions_copy, results=results
            )

            dataset.positions = positions

        except OSError:
            pass

    def modify_settings(self, dataset, results):

        positions_threshold = self.updated_positions_threshold_from_positions(
            positions=dataset.positions, results=results
        )
        self.settings.settings_lens = self.settings.settings_lens.modify_positions_threshold(
            positions_threshold=positions_threshold
        )
        self.check_positions(positions=dataset.positions)

        preload_sparse_grids_of_planes = self.preload_pixelization_grids_of_planes_from_results(
            results=results
        )
        self.settings.settings_pixelization = self.settings.settings_pixelization.modify_preload(
            preload_sparse_grids_of_planes=preload_sparse_grids_of_planes
        )

    def updated_positions_from_positions_and_results(self, positions, results):
        """If automatic position updating is on, update the phase's positions using the results of the previous phase's
        lens model, by ray-tracing backwards the best-fit source centre(s) to the image-plane.

        The outcome of this function are as follows:

        1) If auto positioning is off (self.auto_positions_factor is None), use the previous phase's positions.
        2) If auto positioning is on (self.auto_positions_factor not None) use positions based on the previous phase's
           best-fit tracer. However, if this tracer gives 1 or less positions, use the previous positions.
        3) If this previous tracer is composed of only 1 plane (e.g. you are light profile fitting the image-plane
           only), use the previous positions.
        4) If auto positioning is on or off and there is no previous phase, use the input positions.
        """

        if results.last is not None:
            if not hasattr(results.last, "positions"):
                return positions
            try:
                if len(results.last.max_log_likelihood_tracer.planes) <= 1:
                    return positions
            except AttributeError:
                pass

        if (
            self.settings.settings_lens.auto_positions_factor is not None
            and results.last is not None
        ):

            updated_positions = (
                results.last.image_plane_multiple_image_positions_of_source_plane_centres
            )

            # TODO : Coorrdinates refascotr will sort out index call here

            if isinstance(updated_positions, aa.GridCoordinates):
                if updated_positions.in_list:
                    if len(updated_positions.in_list[0]) > 1:
                        return updated_positions

        if results.last is not None:
            if results.last.positions is not None:
                if results.last.positions.in_list:
                    return results.last.positions

        return positions

    def updated_positions_threshold_from_positions(self, positions, results) -> [float]:
        """
        If automatic position updating is on, update the phase's threshold using this phase's updated positions.

        First, we ray-trace forward the positions of the source-plane centres (see above) via the mass model to
        determine how far apart they are separated. This gives us their source-plane sepration, which is multiplied by
        self.auto_positions_factor to set the threshold.

        The threshold is rounded up to the auto positions minimum threshold if that setting is included."""

        if (
            self.settings.settings_lens.auto_positions_factor
            and results.last is not None
        ):

            if positions is None:
                return None

            positions_fits = fit_positions.FitPositionsSourcePlaneMaxSeparation(
                positions=aa.GridCoordinates(coordinates=positions),
                tracer=results.last.max_log_likelihood_tracer,
                noise_value=1.0,
            )

            positions_threshold = (
                self.settings.settings_lens.auto_positions_factor
                * np.max(positions_fits.maximum_separations)
            )

        else:

            positions_threshold = self.settings.settings_lens.positions_threshold

        if (
            self.settings.settings_lens.auto_positions_minimum_threshold is not None
            and positions_threshold is not None
        ):
            if (
                positions_threshold
                < self.settings.settings_lens.auto_positions_minimum_threshold
            ) or (positions_threshold is None):
                positions_threshold = (
                    self.settings.settings_lens.auto_positions_minimum_threshold
                )

        return positions_threshold

    def preload_pixelization_grids_of_planes_from_results(self, results):

        if self.is_hyper_phase:
            return None

        if (
            results.last is not None
            and self.pixelization is not None
            and not self.pixelization_is_model
        ):
            if self.pixelization.__class__ is results.last.pixelization.__class__:
                if hasattr(results.last, "hyper_combined"):
                    return (
                        results.last.hyper_combined.max_log_likelihood_pixelization_grids_of_planes
                    )
                else:
                    return results.last.max_log_likelihood_pixelization_grids_of_planes
        return None

    def check_positions(self, positions):

        if (
            self.settings.settings_lens.positions_threshold is not None
            and positions is None
        ):
            raise exc.PhaseException(
                "You have specified for a phase to use positions, but not input positions to the "
                "pipeline when you ran it."
            )



from abc import ABC
import pickle
import numpy as np
from astropy import cosmology as cosmo
from autoarray.inversion import pixelizations as pix
from autoarray.inversion import inversions as inv
from autoarray.structures import grids
from autogalaxy import lensing
from autogalaxy.galaxy import galaxy as g
from autogalaxy.plane import plane as pl
from autogalaxy.util import cosmology_util
from autogalaxy.util import plane_util


class AbstractTracer(lensing.LensingObject, ABC):
    def __init__(self, planes, cosmology):
        """Ray-tracer for a lens system with any number of planes.

        The redshift of these planes are specified by the redshits of the galaxies; there is a unique plane redshift \
        for every unique galaxy redshift (galaxies with identical redshifts are put in the same plane).

        To perform multi-plane ray-tracing, a cosmology must be supplied so that deflection-angles can be rescaled \
        according to the lens-geometry of the multi-plane system. All galaxies input to the tracer must therefore \
        have redshifts.

        This tracer has only one grid (see gridStack) which is used for ray-tracing.

        Parameters
        ----------
        galaxies : [Galaxy]
            The list of galaxies in the ray-tracing calculation.
        image_plane_grid : grid_stacks.GridStack
            The image-plane grid which is traced. (includes the grid, sub-grid, blurring-grid, etc.).
        border : masks.GridBorder
            The border of the grid, which is used to relocate demagnified traced pixels to the \
            source-plane borders.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """
        self.planes = planes
        self.plane_redshifts = [plane.redshift for plane in planes]
        self.cosmology = cosmology

    @property
    def total_planes(self):
        return len(self.plane_redshifts)

    @property
    def image_plane(self):
        return self.planes[0]

    @property
    def source_plane(self):
        return self.planes[-1]

    @property
    def galaxies(self):
        return list([galaxy for plane in self.planes for galaxy in plane.galaxies])

    @property
    def all_planes_have_redshifts(self):
        return None not in self.plane_redshifts

    @property
    def has_light_profile(self):
        return any(list(map(lambda plane: plane.has_light_profile, self.planes)))

    @property
    def has_mass_profile(self):
        return any(list(map(lambda plane: plane.has_mass_profile, self.planes)))

    @property
    def has_pixelization(self):
        return any(list(map(lambda plane: plane.has_pixelization, self.planes)))

    @property
    def has_regularization(self):
        return any(list(map(lambda plane: plane.has_regularization, self.planes)))

    @property
    def has_hyper_galaxy(self):
        return any(list(map(lambda plane: plane.has_hyper_galaxy, self.planes)))

    @property
    def upper_plane_index_with_light_profile(self):
        return max(
            [
                plane_index if plane.has_light_profile else 0
                for (plane_index, plane) in enumerate(self.planes)
            ]
        )

    @property
    def planes_with_light_profile(self):
        return list(filter(lambda plane: plane.has_light_profile, self.planes))

    @property
    def planes_with_mass_profile(self):
        return list(filter(lambda plane: plane.has_mass_profile, self.planes))

    @property
    def light_profile_centres(self):
        """Returns the light profile centres of the tracer as a *GridCoordinates* object, which structures the centres
        in lists according to which plane they come from.

        Fo example, if the tracer has two planes, the first with one light profile and second with two light profiles
        this returns:

        [[(y0, x0)], [(y0, x0), (y1, x1)]]

        This is used for visualization, for example plotting the centres of all light profiles colored by their galaxy.

        The centres of light-sheets are filtered out, as their centres are not relevant to lensing calculations

        """
        return grids.GridCoordinates(
            [
                list(plane.light_profile_centres)
                for plane in self.planes
                if plane.has_light_profile
            ]
        )

    @property
    def mass_profiles(self):
        return [
            item
            for mass_profile in self.mass_profiles_of_planes
            for item in mass_profile
        ]

    @property
    def mass_profiles_of_planes(self):
        return [plane.mass_profiles for plane in self.planes if plane.has_mass_profile]

    @property
    def mass_profile_centres(self):
        """Returns the mass profile centres of the tracer as a *GridCoordinates* object, which structures the centres
        in lists according to which plane they come from.

        Fo example, if the tracer has two planes, the first with one mass profile and second with two mass profiles
        this returns:

        [[(y0, x0)], [(y0, x0), (y1, x1)]]

        This is used for visualization, for example plotting the centres of all mass profiles colored by their galaxy.

        The centres of mass-sheets are filtered out, as their centres are not relevant to lensing calculations

        """
        return grids.GridCoordinates(
            [
                list(plane.mass_profile_centres)
                for plane in self.planes
                if plane.has_mass_profile
            ]
        )

    @property
    def plane_indexes_with_pixelizations(self):
        plane_indexes_with_inversions = [
            plane_index if plane.has_pixelization else None
            for (plane_index, plane) in enumerate(self.planes)
        ]
        return [
            plane_index
            for plane_index in plane_indexes_with_inversions
            if plane_index is not None
        ]

    @property
    def pixelizations_of_planes(self):
        return [plane.pixelization for plane in self.planes]

    @property
    def regularizations_of_planes(self):
        return [plane.regularization for plane in self.planes]

    @property
    def hyper_galaxy_image_of_planes_with_pixelizations(self):
        return [
            plane.hyper_galaxy_image_of_galaxy_with_pixelization
            for plane in self.planes
        ]

    def plane_with_galaxy(self, galaxy):
        return [plane for plane in self.planes if galaxy in plane.galaxies][0]

    def new_object_with_units_converted(
        self,
        unit_length=None,
        unit_luminosity=None,
        unit_mass=None,
        kpc_per_arcsec=None,
        exposure_time=None,
        critical_surface_density=None,
    ):

        new_planes = list(
            map(
                lambda plane: plane.new_object_with_units_converted(
                    unit_length=unit_length,
                    unit_luminosity=unit_luminosity,
                    unit_mass=unit_mass,
                    kpc_per_arcsec=kpc_per_arcsec,
                    exposure_time=exposure_time,
                    critical_surface_density=critical_surface_density,
                ),
                self.planes,
            )
        )

        return self.__class__(planes=new_planes, cosmology=self.cosmology)

    @property
    def unit_length(self):
        if self.has_light_profile:
            return self.planes_with_light_profile[0].unit_length
        elif self.has_mass_profile:
            return self.planes_with_mass_profile[0].unit_length
        else:
            return None

    @property
    def unit_luminosity(self):
        if self.has_light_profile:
            return self.planes_with_light_profile[0].unit_luminosity
        elif self.has_mass_profile:
            return self.planes_with_mass_profile[0].unit_luminosity
        else:
            return None

    @property
    def unit_mass(self):
        if self.has_mass_profile:
            return self.planes_with_mass_profile[0].unit_mass
        else:
            return None

    @classmethod
    def load(cls, file_path, filename="tracer"):
        with open(f"{file_path}/{filename}.pickle", "rb") as f:
            return pickle.load(f)

    def save(self, file_path, filename="tracer"):
        """
        Save the tracer by serializing it with pickle.
        """
        with open(f"{file_path}/{filename}.pickle", "wb") as f:
            pickle.dump(self, f)


class AbstractTracerCosmology(AbstractTracer, ABC):
    @property
    def arcsec_per_kpc(self):
        return self.arcsec_per_kpc_proper_of_plane(i=0)

    @property
    def kpc_per_arcsec(self):
        return 1.0 / self.arcsec_per_kpc

    def arcsec_per_kpc_proper_of_plane(self, i):
        return cosmology_util.arcsec_per_kpc_from(
            redshift=self.plane_redshifts[i], cosmology=self.cosmology
        )

    def kpc_per_arcsec_proper_of_plane(self, i):
        return 1.0 / self.arcsec_per_kpc_proper_of_plane(i=i)

    def angular_diameter_distance_of_plane_to_earth_in_units(
        self, i, unit_length="arcsec"
    ):
        return cosmology_util.angular_diameter_distance_to_earth_from(
            redshift=self.plane_redshifts[i],
            cosmology=self.cosmology,
            unit_length=unit_length,
        )

    def angular_diameter_distance_between_planes_in_units(
        self, i, j, unit_length="arcsec"
    ):
        return cosmology_util.angular_diameter_distance_between_redshifts_from(
            redshift_0=self.plane_redshifts[i],
            redshift_1=self.plane_redshifts[j],
            cosmology=self.cosmology,
            unit_length=unit_length,
        )

    def angular_diameter_distance_to_source_plane_in_units(self, unit_length="arcsec"):
        return cosmology_util.angular_diameter_distance_to_earth_from(
            redshift=self.plane_redshifts[-1],
            cosmology=self.cosmology,
            unit_length=unit_length,
        )

    def critical_surface_density_between_planes_in_units(
        self, i, j, unit_length="arcsec", unit_mass="solMass"
    ):
        return cosmology_util.critical_surface_density_between_redshifts_from(
            redshift_0=self.plane_redshifts[i],
            redshift_1=self.plane_redshifts[j],
            cosmology=self.cosmology,
            unit_length=unit_length,
            unit_mass=unit_mass,
        )

    def scaling_factor_between_planes(self, i, j):
        return cosmology_util.scaling_factor_between_redshifts_from(
            redshift_0=self.plane_redshifts[i],
            redshift_1=self.plane_redshifts[j],
            redshift_final=self.plane_redshifts[-1],
            cosmology=self.cosmology,
        )

    def angular_diameter_distance_from_image_to_source_plane_in_units(
        self, unit_length="arcsec"
    ):
        return self.angular_diameter_distance_between_planes_in_units(
            i=0, j=-1, unit_length=unit_length
        )


class AbstractTracerLensing(AbstractTracerCosmology, ABC):
    @grids.grid_like_to_structure_list
    def traced_grids_of_planes_from_grid(self, grid, plane_index_limit=None):

        traced_grids = []
        traced_deflections = []

        for (plane_index, plane) in enumerate(self.planes):

            scaled_grid = grid.copy()

            if plane_index > 0:
                for previous_plane_index in range(plane_index):
                    scaling_factor = cosmology_util.scaling_factor_between_redshifts_from(
                        redshift_0=self.plane_redshifts[previous_plane_index],
                        redshift_1=plane.redshift,
                        redshift_final=self.plane_redshifts[-1],
                        cosmology=self.cosmology,
                    )

                    scaled_deflections = (
                        scaling_factor * traced_deflections[previous_plane_index]
                    )

                    # TODO : Setup as GridInterpolate

                    scaled_grid -= scaled_deflections

            traced_grids.append(scaled_grid)

            if plane_index_limit is not None:
                if plane_index == plane_index_limit:
                    return traced_grids

            traced_deflections.append(plane.deflections_from_grid(grid=scaled_grid))

        return traced_grids

    @grids.grid_like_to_structure
    def deflections_between_planes_from_grid(self, grid, plane_i=0, plane_j=-1):

        traced_grids_of_planes = self.traced_grids_of_planes_from_grid(grid=grid)

        return traced_grids_of_planes[plane_i] - traced_grids_of_planes[plane_j]

    @grids.grid_like_to_structure
    def image_from_grid(self, grid):
        return sum(self.images_of_planes_from_grid(grid=grid))

    @grids.grid_like_to_structure_list
    def images_of_planes_from_grid(self, grid):

        traced_grids_of_planes = self.traced_grids_of_planes_from_grid(
            grid=grid, plane_index_limit=self.upper_plane_index_with_light_profile
        )

        images_of_planes = [
            self.planes[plane_index].image_from_grid(
                grid=traced_grids_of_planes[plane_index]
            )
            for plane_index in range(len(traced_grids_of_planes))
        ]

        if self.upper_plane_index_with_light_profile < self.total_planes - 1:
            for plane_index in range(
                self.upper_plane_index_with_light_profile, self.total_planes - 1
            ):
                images_of_planes.append(np.zeros(shape=images_of_planes[0].shape))

        return images_of_planes

    def padded_image_from_grid_and_psf_shape(self, grid, psf_shape_2d):

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_2d=psf_shape_2d)

        return self.image_from_grid(grid=padded_grid)

    @grids.grid_like_to_structure
    def convergence_from_grid(self, grid):
        return sum([plane.convergence_from_grid(grid=grid) for plane in self.planes])

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        return sum([plane.potential_from_grid(grid=grid) for plane in self.planes])

    @grids.grid_like_to_structure
    def deflections_from_grid(self, grid):
        return self.deflections_between_planes_from_grid(grid=grid)

    @grids.grid_like_to_structure
    def deflections_of_planes_summed_from_grid(self, grid):
        return sum([plane.deflections_from_grid(grid=grid) for plane in self.planes])

    def grid_at_redshift_from_grid_and_redshift(self, grid, redshift):
        """For an input grid of (y,x) arc-second image-plane coordinates, ray-trace the coordinates to any redshift in \
        the strong lens configuration.

        This is performed using multi-plane ray-tracing and the existing redshifts and planes of the tracer. However, \
        any redshift can be input even if a plane does not exist there, including redshifts before the first plane \
        of the lens system.

        Parameters
        ----------
        grid : ndsrray or aa.Grid
            The image-plane grid which is traced to the redshift.
        redshift : float
            The redshift the image-plane grid is traced to.
        """

        if redshift <= self.plane_redshifts[0]:
            return grid.copy()

        plane_index_with_redshift = [
            plane_index
            for plane_index, plane in enumerate(self.planes)
            if plane.redshift == redshift
        ]

        if plane_index_with_redshift:
            return self.traced_grids_of_planes_from_grid(grid=grid)[
                plane_index_with_redshift[0]
            ]

        for plane_index, plane_redshift in enumerate(self.plane_redshifts):

            if redshift < plane_redshift:
                plane_index_insert = plane_index

        planes = self.planes
        planes.insert(
            plane_index_insert,
            pl.Plane(redshift=redshift, galaxies=[], cosmology=self.cosmology),
        )

        tracer = Tracer(planes=planes, cosmology=self.cosmology)

        return tracer.traced_grids_of_planes_from_grid(grid=grid)[plane_index_insert]

    @property
    def contribution_map(self):

        contribution_maps = self.contribution_maps_of_planes
        if None in contribution_maps:
            contribution_maps = [i for i in contribution_maps if i is not None]

        if contribution_maps:
            return sum(contribution_maps)
        else:
            return None

    @property
    def contribution_maps_of_planes(self):

        contribution_maps = []

        for plane in self.planes:

            if plane.contribution_map is not None:

                contribution_maps.append(plane.contribution_map)

            else:

                contribution_maps.append(None)

        return contribution_maps


class AbstractTracerData(AbstractTracerLensing, ABC):
    def blurred_image_from_grid_and_psf(self, grid, psf, blurring_grid):
        """Extract the 1D image and 1D blurring image of every plane and blur each with the \
        PSF using a psf (see imaging.convolution).

        These are summed to give the tracer's overall blurred image in 1D.

        Parameters
        ----------
        psf : hyper_galaxies.imaging.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """

        if not self.has_light_profile:
            return np.zeros(shape=grid.shape_1d)

        image = self.image_from_grid(grid=grid)

        blurring_image = self.image_from_grid(grid=blurring_grid)

        return psf.convolved_array_from_array_2d_and_mask(
            array_2d=image.in_2d_binned + blurring_image.in_2d_binned, mask=grid.mask
        )

    def blurred_images_of_planes_from_grid_and_psf(self, grid, psf, blurring_grid):
        """Extract the 1D image and 1D blurring image of every plane and blur each with the \
        PSF using a psf (see imaging.convolution).

        The blurred image of every plane is returned in 1D.

        Parameters
        ----------
        psf : hyper_galaxies.imaging.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """

        traced_grids_of_planes = self.traced_grids_of_planes_from_grid(grid=grid)
        traced_blurring_grids_of_planes = self.traced_grids_of_planes_from_grid(
            grid=blurring_grid
        )
        return [
            plane.blurred_image_from_grid_and_psf(
                grid=traced_grids_of_planes[plane_index],
                psf=psf,
                blurring_grid=traced_blurring_grids_of_planes[plane_index],
            )
            for (plane_index, plane) in enumerate(self.planes)
        ]

    def blurred_image_from_grid_and_convolver(self, grid, convolver, blurring_grid):
        """Extract the 1D image and 1D blurring image of every plane and blur each with the \
        PSF using a convolver (see imaging.convolution).

        These are summed to give the tracer's overall blurred image in 1D.

        Parameters
        ----------
        convolver : hyper_galaxies.imaging.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """

        if not self.has_light_profile:
            return np.zeros(shape=grid.shape_1d)

        image = self.image_from_grid(grid=grid)

        blurring_image = self.image_from_grid(grid=blurring_grid)

        return convolver.convolved_image_from_image_and_blurring_image(
            image=image, blurring_image=blurring_image
        )

    def blurred_images_of_planes_from_grid_and_convolver(
        self, grid, convolver, blurring_grid
    ):
        """Extract the 1D image and 1D blurring image of every plane and blur each with the \
        PSF using a convolver (see imaging.convolution).

        The blurred image of every plane is returned in 1D.

        Parameters
        ----------
        convolver : hyper_galaxies.imaging.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """

        traced_grids_of_planes = self.traced_grids_of_planes_from_grid(grid=grid)
        traced_blurring_grids_of_planes = self.traced_grids_of_planes_from_grid(
            grid=blurring_grid
        )

        return [
            plane.blurred_image_from_grid_and_convolver(
                grid=traced_grids_of_planes[plane_index],
                convolver=convolver,
                blurring_grid=traced_blurring_grids_of_planes[plane_index],
            )
            for (plane_index, plane) in enumerate(self.planes)
        ]

    def unmasked_blurred_image_from_grid_and_psf(self, grid, psf):

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_2d=psf.shape_2d)

        padded_image = self.image_from_grid(grid=padded_grid)

        return padded_grid.mask.unmasked_blurred_array_from_padded_array_psf_and_image_shape(
            padded_array=padded_image, psf=psf, image_shape=grid.mask.shape
        )

    def unmasked_blurred_image_of_planes_from_grid_and_psf(self, grid, psf):

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_2d=psf.shape_2d)

        traced_padded_grids = self.traced_grids_of_planes_from_grid(grid=padded_grid)

        unmasked_blurred_images_of_planes = []

        for plane, traced_padded_grid in zip(self.planes, traced_padded_grids):
            padded_image_1d = plane.image_from_grid(grid=traced_padded_grid)

            unmasked_blurred_array_2d = padded_grid.mask.unmasked_blurred_array_from_padded_array_psf_and_image_shape(
                padded_array=padded_image_1d, psf=psf, image_shape=grid.mask.shape
            )

            unmasked_blurred_images_of_planes.append(unmasked_blurred_array_2d)

        return unmasked_blurred_images_of_planes

    def unmasked_blurred_image_of_planes_and_galaxies_from_grid_and_psf(
        self, grid, psf
    ):

        unmasked_blurred_images_of_planes_and_galaxies = []

        padded_grid = grid.padded_grid_from_kernel_shape(kernel_shape_2d=psf.shape_2d)

        traced_padded_grids = self.traced_grids_of_planes_from_grid(grid=padded_grid)

        for plane, traced_padded_grid in zip(self.planes, traced_padded_grids):
            padded_image_1d_of_galaxies = plane.images_of_galaxies_from_grid(
                grid=traced_padded_grid
            )

            unmasked_blurred_array_2d_of_galaxies = list(
                map(
                    lambda padded_image_1d_of_galaxy: padded_grid.mask.unmasked_blurred_array_from_padded_array_psf_and_image_shape(
                        padded_array=padded_image_1d_of_galaxy,
                        psf=psf,
                        image_shape=grid.mask.shape,
                    ),
                    padded_image_1d_of_galaxies,
                )
            )

            unmasked_blurred_images_of_planes_and_galaxies.append(
                unmasked_blurred_array_2d_of_galaxies
            )

        return unmasked_blurred_images_of_planes_and_galaxies

    def profile_visibilities_from_grid_and_transformer(self, grid, transformer):

        if not self.has_light_profile:
            return np.zeros(shape=transformer.uv_wavelengths.shape)

        image = self.image_from_grid(grid=grid)

        return transformer.visibilities_from_image(image=image)

    def profile_visibilities_of_planes_from_grid_and_transformer(
        self, grid, transformer
    ):

        images_of_planes = self.images_of_planes_from_grid(grid=grid)
        return [
            transformer.visibilities_from_image(image=image)
            for image in images_of_planes
        ]

    def sparse_image_plane_grids_of_planes_from_grid(
        self, grid, pixelization_setting=pix.SettingsPixelization()
    ):

        sparse_image_plane_grids_of_planes = []

        for plane in self.planes:
            sparse_image_plane_grid = plane.sparse_image_plane_grid_from_grid(
                grid=grid, settings_pixelization=pixelization_setting
            )
            sparse_image_plane_grids_of_planes.append(sparse_image_plane_grid)

        return sparse_image_plane_grids_of_planes

    def traced_sparse_grids_of_planes_from_grid(
        self, grid, settings_pixelization=pix.SettingsPixelization()
    ):

        if (
            settings_pixelization.preload_sparse_grids_of_planes is None
            or settings_pixelization.is_stochastic
        ):

            sparse_image_plane_grids_of_planes = self.sparse_image_plane_grids_of_planes_from_grid(
                grid=grid, pixelization_setting=settings_pixelization
            )

        else:

            sparse_image_plane_grids_of_planes = (
                settings_pixelization.preload_sparse_grids_of_planes
            )

        traced_sparse_grids_of_planes = []

        for (plane_index, plane) in enumerate(self.planes):

            if sparse_image_plane_grids_of_planes[plane_index] is None:
                traced_sparse_grids_of_planes.append(None)
            else:
                traced_sparse_grids = self.traced_grids_of_planes_from_grid(
                    grid=sparse_image_plane_grids_of_planes[plane_index]
                )
                traced_sparse_grids_of_planes.append(traced_sparse_grids[plane_index])

        return traced_sparse_grids_of_planes

    def mappers_of_planes_from_grid(
        self, grid, settings_pixelization=pix.SettingsPixelization()
    ):

        mappers_of_planes = []

        traced_grids_of_planes = self.traced_grids_of_planes_from_grid(grid=grid)

        traced_sparse_grids_of_planes = self.traced_sparse_grids_of_planes_from_grid(
            grid=grid, settings_pixelization=settings_pixelization
        )

        for (plane_index, plane) in enumerate(self.planes):

            if not plane.has_pixelization:
                mappers_of_planes.append(None)
            else:
                mapper = plane.mapper_from_grid_and_sparse_grid(
                    grid=traced_grids_of_planes[plane_index],
                    sparse_grid=traced_sparse_grids_of_planes[plane_index],
                    settings_pixelization=settings_pixelization,
                )
                mappers_of_planes.append(mapper)

        return mappers_of_planes

    def inversion_imaging_from_grid_and_data(
        self,
        grid,
        image,
        noise_map,
        convolver,
        settings_pixelization=pix.SettingsPixelization(),
        settings_inversion=inv.SettingsInversion(),
    ):

        mappers_of_planes = self.mappers_of_planes_from_grid(
            grid=grid, settings_pixelization=settings_pixelization
        )

        return inv.InversionImagingMatrix.from_data_mapper_and_regularization(
            image=image,
            noise_map=noise_map,
            convolver=convolver,
            mapper=mappers_of_planes[-1],
            regularization=self.regularizations_of_planes[-1],
            settings=settings_inversion,
        )

    def inversion_interferometer_from_grid_and_data(
        self,
        grid,
        visibilities,
        noise_map,
        transformer,
        settings_pixelization=pix.SettingsPixelization(),
        settings_inversion=inv.SettingsInversion(),
    ):
        mappers_of_planes = self.mappers_of_planes_from_grid(
            grid=grid, settings_pixelization=settings_pixelization
        )

        return inv.AbstractInversionInterferometer.from_data_mapper_and_regularization(
            visibilities=visibilities,
            noise_map=noise_map,
            transformer=transformer,
            mapper=mappers_of_planes[-1],
            regularization=self.regularizations_of_planes[-1],
            settings=settings_inversion,
        )

    def hyper_noise_map_from_noise_map(self, noise_map):
        hyper_noise_maps = self.hyper_noise_maps_of_planes_from_noise_map(
            noise_map=noise_map
        )
        return sum(hyper_noise_maps)

    def hyper_noise_maps_of_planes_from_noise_map(self, noise_map):
        return [
            plane.hyper_noise_map_from_noise_map(noise_map=noise_map)
            for plane in self.planes
        ]

    def galaxy_image_dict_from_grid(self, grid) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_image_dict = dict()

        traced_grids_of_planes = self.traced_grids_of_planes_from_grid(grid=grid)

        for (plane_index, plane) in enumerate(self.planes):
            images_of_galaxies = plane.images_of_galaxies_from_grid(
                grid=traced_grids_of_planes[plane_index]
            )
            for (galaxy_index, galaxy) in enumerate(plane.galaxies):
                galaxy_image_dict[galaxy] = images_of_galaxies[galaxy_index]

        return galaxy_image_dict

    def galaxy_blurred_image_dict_from_grid_and_convolver(
        self, grid, convolver, blurring_grid
    ) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_blurred_image_dict = dict()

        traced_grids_of_planes = self.traced_grids_of_planes_from_grid(grid=grid)

        traced_blurring_grids_of_planes = self.traced_grids_of_planes_from_grid(
            grid=blurring_grid
        )

        for (plane_index, plane) in enumerate(self.planes):
            blurred_images_of_galaxies = plane.blurred_images_of_galaxies_from_grid_and_convolver(
                grid=traced_grids_of_planes[plane_index],
                convolver=convolver,
                blurring_grid=traced_blurring_grids_of_planes[plane_index],
            )
            for (galaxy_index, galaxy) in enumerate(plane.galaxies):
                galaxy_blurred_image_dict[galaxy] = blurred_images_of_galaxies[
                    galaxy_index
                ]

        return galaxy_blurred_image_dict

    def galaxy_profile_visibilities_dict_from_grid_and_transformer(
        self, grid, transformer
    ) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_profile_visibilities_image_dict = dict()

        traced_grids_of_planes = self.traced_grids_of_planes_from_grid(grid=grid)

        for (plane_index, plane) in enumerate(self.planes):
            profile_visibilities_of_galaxies = plane.profile_visibilities_of_galaxies_from_grid_and_transformer(
                grid=traced_grids_of_planes[plane_index], transformer=transformer
            )
            for (galaxy_index, galaxy) in enumerate(plane.galaxies):
                galaxy_profile_visibilities_image_dict[
                    galaxy
                ] = profile_visibilities_of_galaxies[galaxy_index]

        return galaxy_profile_visibilities_image_dict


class Tracer(AbstractTracerData):
    @classmethod
    def from_galaxies(cls, galaxies, cosmology=cosmo.Planck15):

        plane_redshifts = plane_util.ordered_plane_redshifts_from(galaxies=galaxies)

        galaxies_in_planes = plane_util.galaxies_in_redshift_ordered_planes_from(
            galaxies=galaxies, plane_redshifts=plane_redshifts
        )

        planes = []

        for plane_index in range(0, len(plane_redshifts)):
            planes.append(
                pl.Plane(galaxies=galaxies_in_planes[plane_index], cosmology=cosmology)
            )

        return Tracer(planes=planes, cosmology=cosmology)

    @classmethod
    def sliced_tracer_from_lens_line_of_sight_and_source_galaxies(
        cls,
        lens_galaxies,
        line_of_sight_galaxies,
        source_galaxies,
        planes_between_lenses,
        cosmology=cosmo.Planck15,
    ):

        """Ray-tracer for a lens system with any number of planes.

        The redshift of these planes are specified by the input parameters *lens_redshifts* and \
         *slices_between_main_planes*. Every galaxy is placed in its closest plane in redshift-space.

        To perform multi-plane ray-tracing, a cosmology must be supplied so that deflection-angles can be rescaled \
        according to the lens-geometry of the multi-plane system. All galaxies input to the tracer must therefore \
        have redshifts.

        This tracer has only one grid (see gridStack) which is used for ray-tracing.

        Parameters
        ----------
        lens_galaxies : [Galaxy]
            The list of galaxies in the ray-tracing calculation.
        image_plane_grid : grid_stacks.GridStack
            The image-plane grid which is traced. (includes the grid, sub-grid, blurring-grid, etc.).
        planes_between_lenses : [int]
            The number of slices between each main plane. The first entry in this list determines the number of slices \
            between Earth (redshift 0.0) and main plane 0, the next between main planes 0 and 1, etc.
        border : masks.GridBorder
            The border of the grid, which is used to relocate demagnified traced pixels to the \
            source-plane borders.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        lens_redshifts = plane_util.ordered_plane_redshifts_from(galaxies=lens_galaxies)

        plane_redshifts = plane_util.ordered_plane_redshifts_with_slicing_from(
            lens_redshifts=lens_redshifts,
            planes_between_lenses=planes_between_lenses,
            source_plane_redshift=source_galaxies[0].redshift,
        )

        galaxies_in_planes = plane_util.galaxies_in_redshift_ordered_planes_from(
            galaxies=lens_galaxies + line_of_sight_galaxies,
            plane_redshifts=plane_redshifts,
        )

        plane_redshifts.append(source_galaxies[0].redshift)
        galaxies_in_planes.append(source_galaxies)

        planes = []

        for plane_index in range(0, len(plane_redshifts)):
            planes.append(
                pl.Plane(
                    redshift=plane_redshifts[plane_index],
                    galaxies=galaxies_in_planes[plane_index],
                    cosmology=cosmology,
                )
            )

        return Tracer(planes=planes, cosmology=cosmology)
